@@ -14,13 +14,107 @@
 
 """A library of random helper functionality."""
 
-import platform, subprocess, operator, os, shutil, re, sys
+import stat
+import platform, subprocess, operator, os, shutil, re
 
 from glob import glob
 
 class MesonException(Exception):
-    def __init__(self, *args, **kwargs):
-        Exception.__init__(self, *args, **kwargs)
+    '''Exceptions thrown by Meson'''
+
+class EnvironmentException(MesonException):
+    '''Exceptions thrown while processing and creating the build environment'''
+
+class FileMode:
+    # The first triad is for owner permissions, the second for group permissions,
+    # and the third for others (everyone else).
+    # For the 1st character:
+    #  'r' means can read
+    #  '-' means not allowed
+    # For the 2nd character:
+    #  'w' means can write
+    #  '-' means not allowed
+    # For the 3rd character:
+    #  'x' means can execute
+    #  's' means can execute and setuid/setgid is set (owner/group triads only)
+    #  'S' means cannot execute and setuid/setgid is set (owner/group triads only)
+    #  't' means can execute and sticky bit is set ("others" triads only)
+    #  'T' means cannot execute and sticky bit is set ("others" triads only)
+    #  '-' means none of these are allowed
+    #
+    # The meanings of 'rwx' perms is not obvious for directories; see:
+    # https://www.hackinglinuxexposed.com/articles/20030424.html
+    #
+    # For information on this notation such as setuid/setgid/sticky bits, see:
+    # https://en.wikipedia.org/wiki/File_system_permissions#Symbolic_notation
+    symbolic_perms_regex = re.compile('[r-][w-][xsS-]' # Owner perms
+                                      '[r-][w-][xsS-]' # Group perms
+                                      '[r-][w-][xtT-]') # Others perms
+
+    def __init__(self, perms=None, owner=None, group=None):
+        self.perms_s = perms
+        self.perms = self.perms_s_to_bits(perms)
+        self.owner = owner
+        self.group = group
+
+    def __repr__(self):
+        ret = '<FileMode: {!r} owner={} group={}'
+        return ret.format(self.perms_s, self.owner, self.group)
+
+    @classmethod
+    def perms_s_to_bits(cls, perms_s):
+        '''
+        Does the opposite of stat.filemode(), converts strings of the form
+        'rwxr-xr-x' to st_mode enums which can be passed to os.chmod()
+        '''
+        if perms_s is None:
+            # No perms specified, we will not touch the permissions
+            return -1
+        eg = 'rwxr-xr-x'
+        if not isinstance(perms_s, str):
+            msg = 'Install perms must be a string. For example, {!r}'
+            raise MesonException(msg.format(eg))
+        if len(perms_s) != 9 or not cls.symbolic_perms_regex.match(perms_s):
+            msg = 'File perms {!r} must be exactly 9 chars. For example, {!r}'
+            raise MesonException(msg.format(perms_s, eg))
+        perms = 0
+        # Owner perms
+        if perms_s[0] == 'r':
+            perms |= stat.S_IRUSR
+        if perms_s[1] == 'w':
+            perms |= stat.S_IWUSR
+        if perms_s[2] == 'x':
+            perms |= stat.S_IXUSR
+        elif perms_s[2] == 'S':
+            perms |= stat.S_ISUID
+        elif perms_s[2] == 's':
+            perms |= stat.S_IXUSR
+            perms |= stat.S_ISUID
+        # Group perms
+        if perms_s[3] == 'r':
+            perms |= stat.S_IRGRP
+        if perms_s[4] == 'w':
+            perms |= stat.S_IWGRP
+        if perms_s[5] == 'x':
+            perms |= stat.S_IXGRP
+        elif perms_s[5] == 'S':
+            perms |= stat.S_ISGID
+        elif perms_s[5] == 's':
+            perms |= stat.S_IXGRP
+            perms |= stat.S_ISGID
+        # Others perms
+        if perms_s[6] == 'r':
+            perms |= stat.S_IROTH
+        if perms_s[7] == 'w':
+            perms |= stat.S_IWOTH
+        if perms_s[8] == 'x':
+            perms |= stat.S_IXOTH
+        elif perms_s[8] == 'T':
+            perms |= stat.S_ISVTX
+        elif perms_s[8] == 't':
+            perms |= stat.S_IXOTH
+            perms |= stat.S_ISVTX
+        return perms
 
 class File:
     def __init__(self, is_built, subdir, fname):
@@ -58,6 +152,12 @@ class File:
         else:
             return os.path.join(build_to_src, self.subdir, self.fname)
 
+    def absolute_path(self, srcdir, builddir):
+        if self.is_built:
+            return os.path.join(builddir, self.subdir, self.fname)
+        else:
+            return os.path.join(srcdir, self.subdir, self.fname)
+
     def endswith(self, ending):
         return self.fname.endswith(ending)
 
@@ -70,6 +170,9 @@ class File:
     def __hash__(self):
         return hash((self.fname, self.subdir, self.is_built))
 
+    def relative_name(self):
+        return os.path.join(self.subdir, self.fname)
+
 def get_compiler_for_source(compilers, src):
     for comp in compilers:
         if comp.can_compile(src):
@@ -137,21 +240,26 @@ def detect_vcs(source_dir):
                 return vcs
     return None
 
-def grab_leading_numbers(vstr):
+def grab_leading_numbers(vstr, strict=False):
     result = []
     for x in vstr.split('.'):
         try:
             result.append(int(x))
-        except ValueError:
+        except ValueError as e:
+            if strict:
+                msg = 'Invalid version to compare against: {!r}; only ' \
+                      'numeric digits separated by "." are allowed: ' + str(e)
+                raise MesonException(msg.format(vstr))
             break
     return result
 
 numpart = re.compile('[0-9.]+')
 
-def version_compare(vstr1, vstr2):
+def version_compare(vstr1, vstr2, strict=False):
     match = numpart.match(vstr1.strip())
     if match is None:
-        raise MesonException('Uncomparable version string %s.' % vstr1)
+        msg = 'Uncomparable version string {!r}.'
+        raise MesonException(msg.format(vstr1))
     vstr1 = match.group(0)
     if vstr2.startswith('>='):
         cmpop = operator.ge
@@ -176,10 +284,22 @@ def version_compare(vstr1, vstr2):
         vstr2 = vstr2[1:]
     else:
         cmpop = operator.eq
-    varr1 = grab_leading_numbers(vstr1)
-    varr2 = grab_leading_numbers(vstr2)
+    varr1 = grab_leading_numbers(vstr1, strict)
+    varr2 = grab_leading_numbers(vstr2, strict)
     return cmpop(varr1, varr2)
 
+def version_compare_many(vstr1, conditions):
+    if not isinstance(conditions, (list, tuple)):
+        conditions = [conditions]
+    found = []
+    not_found = []
+    for req in conditions:
+        if not version_compare(vstr1, req, strict=True):
+            not_found.append(req)
+        else:
+            found.append(req)
+    return not_found == [], not_found, found
+
 def default_libdir():
     if is_debianlike():
         try:
@@ -288,13 +408,13 @@ def do_conf_file(src, dst, confdata):
             line = do_replacement(regex, line, confdata)
         result.append(line)
     dst_tmp = dst + '~'
-    with open(dst_tmp, 'w') as f:
+    with open(dst_tmp, 'w', encoding='utf-8') as f:
         f.writelines(result)
     shutil.copymode(src, dst_tmp)
     replace_if_different(dst, dst_tmp)
 
 def dump_conf_header(ofilename, cdata):
-    with open(ofilename, 'w') as ofile:
+    with open(ofilename, 'w', encoding='utf-8') as ofile:
         ofile.write('''/*
  * Autogenerated by the Meson build system.
  * Do not edit, your changes will be lost.
@@ -332,11 +452,21 @@ def replace_if_different(dst, dst_tmp):
     else:
         os.unlink(dst_tmp)
 
+def stringintlistify(item):
+    if isinstance(item, (str, int)):
+        item = [item]
+    if not isinstance(item, list):
+        raise MesonException('Item must be a list, a string, or an int')
+    for i in item:
+        if not isinstance(i, (str, int, type(None))):
+            raise MesonException('List item must be a string or an int')
+    return item
+
 def stringlistify(item):
     if isinstance(item, str):
         item = [item]
     if not isinstance(item, list):
-        raise MesonException('Item is not an array')
+        raise MesonException('Item is not a list')
     for i in item:
         if not isinstance(i, str):
             raise MesonException('List item not a string.')
@@ -359,3 +489,10 @@ def expand_arguments(args):
             print(e)
             return None
     return expended_args
+
+def Popen_safe(args, write=None, stderr=subprocess.PIPE, **kwargs):
+    p = subprocess.Popen(args, universal_newlines=True,
+                         stdout=subprocess.PIPE,
+                         stderr=stderr, **kwargs)
+    o, e = p.communicate(write)
+    return p, o, e
