@@ -20,10 +20,12 @@ from .. import compilers
 import datetime
 from .. import mlog
 from . import GirTarget, TypelibTarget
+from . import ModuleReturnValue
+from . import ExtensionModule
 
 import os
 
-class RPMModule:
+class RPMModule(ExtensionModule):
 
     def generate_spec_template(self, state, args, kwargs):
         compiler_deps = set()
@@ -55,7 +57,7 @@ class RPMModule:
                 files.add('%%{_bindir}/%s' % target.get_filename())
             elif isinstance(target, build.SharedLibrary) and target.need_install:
                 files.add('%%{_libdir}/%s' % target.get_filename())
-                for alias in target.get_aliaslist():
+                for alias in target.get_aliases():
                     if alias.endswith('.so'):
                         files_devel.add('%%{_libdir}/%s' % alias)
                     else:
@@ -153,6 +155,7 @@ class RPMModule:
             fn.write('- \n')
             fn.write('\n')
         mlog.log('RPM spec template written to %s.spec.\n' % proj)
+        return ModuleReturnValue(None, [])
 
 def initialize():
     return RPMModule()
