@@ -93,9 +93,19 @@ class Resolver:
     def resolve(self, packagename):
         fname = os.path.join(self.subdir_root, packagename + '.wrap')
         dirname = os.path.join(self.subdir_root, packagename)
-        if os.path.isdir(dirname):
-            # The directory is there? Great, use it.
-            return packagename 
+        try:
+            if os.listdir(dirname):
+                # The directory is there and not empty? Great, use it.
+                return packagename
+            else:
+                mlog.warning('Subproject directory %s is empty, possibly because of an unfinished'
+                             'checkout, removing to reclone' % dirname)
+                os.rmdir(dirname)
+        except NotADirectoryError:
+            raise RuntimeError('%s is not a directory, can not use as subproject.' % dirname)
+        except FileNotFoundError:
+            pass
+
         if not os.path.isfile(fname):
             # No wrap file with this name? Give up.
             return None
@@ -118,6 +128,15 @@ class Resolver:
         revno = p.get('revision')
         is_there = os.path.isdir(checkoutdir)
         if is_there:
+            try:
+                subprocess.check_call(['git', 'rev-parse'])
+                is_there = True
+            except subprocess.CalledProcessError:
+                raise RuntimeError('%s is not empty but is not a valid '
+                                   'git repository, we can not work with it'
+                                   ' as a subproject directory.' % (
+                                       checkoutdir))
+
             if revno.lower() == 'head':
                 # Failure to do pull is not a fatal error,
                 # because otherwise you can't develop without
@@ -134,6 +153,12 @@ class Resolver:
             if revno.lower() != 'head':
                 subprocess.check_call(['git', 'checkout', revno],
                                       cwd=checkoutdir)
+            push_url = p.values.get('push-url')
+            if push_url:
+                subprocess.check_call(['git', 'remote', 'set-url',
+                                       '--push', 'origin', push_url],
+                                      cwd=checkoutdir)
+
     def get_hg(self, p):
         checkoutdir = os.path.join(self.subdir_root, p.get('directory'))
         revno = p.get('revision')
@@ -157,7 +182,7 @@ class Resolver:
                                       cwd=checkoutdir)
 
     def get_data(self, url):
-        blocksize = 10*1024
+        blocksize = 10 * 1024
         if url.startswith('https://wrapdb.mesonbuild.com'):
             resp = open_wrapdburl(url)
         else:
@@ -182,7 +207,7 @@ class Resolver:
                     break
                 downloaded += len(block)
                 blocks.append(block)
-                ratio = int(downloaded/dlsize * 10)
+                ratio = int(downloaded / dlsize * 10)
                 while printed_dots < ratio:
                     print('.', end='')
                     sys.stdout.flush()
